#!/bin/bash

print_usage() {
  local RC=$1
  local ME=${0##*/}
  cat <<- EOT
	usage: $ME [-dhist] [-l level]
	    -d          daemonize
	    -h          help
	    -i          initial spindown
	    -l level    log level
	    -s          log to syslog
	    -t          test / one-shot mode
	EOT

  exit $RC
}

msg() {
  local CUR_LEVEL=$1
  local MSG=$2
  local MIN_LEVEL=${G_LOG_LEVEL:-1}
  local SYSLOG=$G_LOG_SYSLOG
  local TXT_LEVEL

  if [[ $CUR_LEVEL -ge $MIN_LEVEL ]]
  then
    [[ 1 -ne $SYSLOG ]] && unset SYSLOG

    case $CUR_LEVEL in
    ( 0 ) TXT_LEVEL=${SYSLOG:-'[DEBUG  ]'}${SYSLOG:+debug}   ;;
    ( 1 ) TXT_LEVEL=${SYSLOG:-'[INFO   ]'}${SYSLOG:+info}    ;;
    ( 2 ) TXT_LEVEL=${SYSLOG:-'[WARN   ]'}${SYSLOG:+warning} ;;
    ( 3 ) TXT_LEVEL=${SYSLOG:-'[ERROR  ]'}${SYSLOG:+err}     ;;
    ( * ) TXT_LEVEL=${SYSLOG:-'[UNKNOWN]'}${SYSLOG:+notice}  ;;
    esac

    shift
    if [[ 1 -eq $SYSLOG ]]
    then
      logger -p local0.$TXT_LEVEL -i -s -t hdd_spindown "$MSG"
    else
      echo "$(get_unixtime) $TXT_LEVEL $MSG" 1>&2
    fi
  fi
}

debug () {
  msg 0 "$1"
}

info () {
  msg 1 "$1"
}

warn () {
  msg 2 "$1"
}

error() {
  msg 3 "$1"
}

query_db () {
  SQLITE_BIN=/usr/local/bin/sqlite3
  SQLITE_DB=/data/freenas-v1.db

  debug "Database will be accessd by '$SQLITE_BIN $SQLITE_DB <STATEMENT>'"
  $SQLITE_BIN $SQLITE_DB
}

get_xfrs_by_dev () {
  local DEV=$1

  iostat -I $DEV               \
  | awk '$4~/[^a-z]/{printf $4}'
}

get_unixtime () {
  date +%s
}

print_sql_devices () {
  cat <<- EOT
	SELECT disk_name,
	       disk_hddstandby
	FROM storage_disk
	WHERE disk_enabled = 1 AND
	      disk_hddstandby != 'Always On';
	EOT
}

update_dev_cfg () {
  local CNT=0
  local TIME_FACTOR=60

  debug "HDD standby time from database will be multiplied by '$TIME_FACTOR' to get seconds"

  while IFS=\| read COL1 COL2
  do
    if [[ ! $COL1 =~ ^[a-z]+[0-9]+$  \
       || ! $COL2 =~ ^[0-9]+$       ]]
    then
      error "Got at leaset one unexpected value from database: COL1=$COL1, COL2=$COL2"
      continue
    else
      debug "Values from database are syntactically correct: COL1=$COL1, COL2=$COL2"
    fi

    DISK_DEV[$CNT]=$COL1
    DISK_TIMEOUT[$CNT]=$(( $COL2 * $TIME_FACTOR ))

    info "Drive ${DISK_DEV[$CNT]} is configured to be spun down after '${DISK_TIMEOUT[$CNT]}' seconds of inactivity"
    ((CNT++))
  done <<- EOT
	$(query_db <<< $(print_sql_devices))
	EOT
}

update_dev_xfer () {
  local CNT=${#DISK_DEV[@]}

  while ((CNT--))
  do
    DISK_XFER_CURR[$CNT]=$(get_xfrs_by_dev ${DISK_DEV[$CNT]})

    debug "Drive '${DISK_DEV[$CNT]}' has '${DISK_XFER_CURR[$CNT]}' transfers"

    if [[ ${DISK_XFER_CURR[$CNT]} != ${DISK_XFER_LAST[$CNT]} ]]
    then
      debug "There was activity on drive since last check '${DISK_XFER_CURR[$CNT]} != ${DISK_XFER_LAST[$CNT]:-NULL}'"

      DISK_XFER_TIME[$CNT]=$(get_unixtime)
      DISK_XFER_LAST[$CNT]=${DISK_XFER_CURR[$CNT]}
      DISK_SPINNING[$CNT]=1

      debug "Updated last transfer time/count for drive '${DISK_DEV[$CNT]}' to '${DISK_XFER_TIME[$CNT]}/${DISK_XFER_LAST[$CNT]}'"
    else
      debug "There was no activity on drive since last check"
    fi
  done
}

spindown_by_idx () {
  local IDX=$1
  local SPINDOWN_CMD=()
  local CNT
  local DEV

  SPINDOWN_CMD+=('/usr/local/sbin/ataidle -s $DEV')
  SPINDOWN_CMD+=('ataidle -s $DEV')
  SPINDOWN_CMD+=('/usr/local/sbin/ataidle -s /dev/$DEV')
  SPINDOWN_CMD+=('ataidle -s /dev/$DEV')
  SPINDOWN_CMD+=('atacontrol spindown $DEV')
  SPINDOWN_CMD+=('camcontrol stop $DEV')

  if [[ 1 -eq ${DISK_IGNORE[$IDX]} ]]
  then
    info "Omit spindown drive '${DISK_DEV[$IDX]}' because it is set to be ignored"
  else
    if [[ 1 -eq ${DISK_SPINNING[$IDX]} ]]
    then
      debug "Drive '${DISK_DEV[$IDX]}' is marked to be spinning"

      if [[ -n ${DISK_CMD[$IDX]} ]]
      then
        debug "Seems we alread evaluated a working command"
        SPINDOWN_CMD+=("${DISK_CMD[$IDX]}")
      fi

      CNT=${#SPINDOWN_CMD[@]}

      while ((CNT--))
      do
	DEV=${DISK_DEV[$IDX]}
        if eval ${SPINDOWN_CMD[$CNT]} &> /dev/null
        then
          DISK_SPINNING[$IDX]=0
          DISK_CMD[$IDX]=${SPINDOWN_CMD[$CNT]}

          info "Drive '${DISK_DEV[$IDX]}' successfully spun down by command '$(eval echo ${SPINDOWN_CMD[$CNT]})'"
          return
        else
          warn "Spinning down drive by command '$(eval echo ${SPINDOWN_CMD[$CNT]})' failed"
        fi
      done

      error "There was no way to spin down drive '${DISK_DEV[$IDX]}'"
      warn "Avoid future testing by setting drive '${DISK_DEV[$IDX]}' to be ignored"

      DISK_IGNORE[$IDX]=1
    else
      debug "Drive '${DISK_DEV[$IDX]}' is marked to be already spun down"
    fi
  fi
}

conditional_spindown () {
  local CNT=${#DISK_DEV[@]}
  local NOW=$(get_unixtime)
  local NO_ACTIVITY

  while ((CNT--))
  do
    NO_ACTIVITY=$(( $NOW - ${DISK_XFER_TIME[$CNT]} ))

    if [[ $NO_ACTIVITY -ge ${DISK_TIMEOUT[$CNT]} || 1 == $INITIAL_SPINDOWN ]]
    then
      if [[ 1 == $INITIAL_SPINDOWN ]]
      then
        debug "Initial spin down of drive '${DISK_DEV[$CNT]}'."
      else
        debug "Spin down condition for drive '${DISK_DEV[$CNT]}' met. No activity since '$NO_ACTIVITY' seconds and drive has a timeout of ${DISK_TIMEOUT[$CNT]}'."
      fi
      spindown_by_idx $CNT
    else
      debug "Spin down condition for drive '${DISK_DEV[$CNT]}' not met. '$NO_ACTIVITY' seconds since last activity but drive has a timeout of '${DISK_TIMEOUT[$CNT]}' seconds."
    fi
  done
}

get_next_check_time () {
  local CNT=${#DISK_DEV[@]}
  local NOW=$(get_unixtime)
  local NXT=0
  local TIMEOUT=0

  while ((CNT--))
  do
    TIMEOUT=$(( ( ${DISK_XFER_TIME[$CNT]} + ${DISK_TIMEOUT[$CNT]} ) - $NOW ))

    if [[ 0 -lt $TIMEOUT ]]
    then
      debug "Drive '${DISK_DEV[$CNT]}' would time out in '$TIMEOUT' seconds."
    else
      debug "Drive '${DISK_DEV[$CNT]}' timed  out since '$(( -1 * $TIMEOUT ))' seconds."
      TIMEOUT=${DISK_TIMEOUT[$CNT]}
    fi

    debug "Next check for drive '${DISK_DEV[$CNT]}' in '$TIMEOUT' seconds"

    [[ 0 -eq $NXT || $NXT -gt $TIMEOUT ]] && NXT=$TIMEOUT
  done

  debug "Next check in '$NXT' seconds"
  echo -n $NXT
}

fork_me () {
  nohup $0 $@ -s > /dev/null 2>&1 &
  exit
}

main () {
  local  VERSION=0.4
  local  REFRESH_CFG=12
  local  NEXT_CHECK
  local  DISK_DEV
  local  DISK_TIMEOUT
  local  DISK_XFER_CURR
  local  DISK_XFER_LAST
  local  DISK_XFER_TIME
  local  DISK_SPINNING
  local  DISK_CMD
  local  DISK_IGNORE
  local  CNT
  local  OPT
  local  OPTARG
  local  FORK_PAR
  local  DAEMONIZE=0
  local  HELP=0
  local  INITIAL_SPINDOWN=0
  local  ONE_SHOT=0
  local  G_LOG_LEVEL=1
  local  G_LOG_SYSLOG=0

  while getopts dhil:st OPT
  do
    case $OPT in
    ( d ) DAEMONIZE=1         ;;
    ( h ) HELP=1              ;;
    ( i ) INITIAL_SPINDOWN=1  ;;
    ( l ) G_LOG_LEVEL=$OPTARG ;;
    ( s ) G_LOG_SYSLOG=1      ;;
    ( t ) ONE_SHOT=1          ;;
    ( * ) print_usage 1       ;;
    esac
  done

  [[ 1 -eq $HELP ]] && print_usage 0

  if [[ $G_LOG_LEVEL =~ ^[0123]$ ]]
  then
    debug "Log level set to '$G_LOG_LEVEL'"
  else
    error "Invalid log level '$G_LOG_LEVEL' passed. Fallback to '1'."
    G_LOG_LEVEL=1
  fi

  if [[ 1 -eq $DAEMONIZE ]]
  then
    FORK_PAR=${@//-d/}
    FORK_PAR=${FORK_PAR//d/}

    debug "Fork an instance and quit"
    fork_me $FORK_PAR
  fi

  debug "${0##*/} version '$VERSION'"

  [[ 0 -eq $ONE_SHOT ]] \
  && info "Configured HDD standby times will be refreshed after every '$REFRESH_CFG' checks"

  [[ 0 -eq $INITIAL_SPINDOWN ]] \
  && info "Initial check does not spin down any drive"

  while :
  do
    CNT=$REFRESH_CFG

    debug "Updating HDD standby times"
    update_dev_cfg

    while ((CNT--))
    do
      debug "Updating HDD transfer information"
      update_dev_xfer
      debug "Spindown drives depending on their last transfer/standby time"
      conditional_spindown

      if [[ 0 -eq $ONE_SHOT ]]
      then 
        NEXT_CHECK=$(get_next_check_time)

        info "Sleeping '$NEXT_CHECK' seconds till next check. '$CNT' checks left to next update of HDD standby times."
        sleep $NEXT_CHECK
      fi
    done

    if [[ 1 -eq $ONE_SHOT ]]
    then
      info "Test/One-shot mode. Exiting."
      break
    fi
  done
}

export G_LOG_LEVEL=1

main "$@"
exit $?
